/**
 * @file
 * @preserve
 * DOM content synchronization with Mustache templates.
 *
 * Using Mustache.js by
 *   Copyright (c) 2009 Chris Wanstrath (Ruby)
 *   Copyright (c) 2010-2014 Jan Lehnardt (JavaScript)
 *   Copyright (c) 2010-2015 The mustache.js community
 */

window.mustacheSync = window.mustacheSync || {items: [], templates: []};

(function (Drupal, sync, Mustache, window) {
  sync.registry = sync.registry || {
    items: [],
    pending: [],
    templates: {},
    providers: {},
    listeners: {}
  };

  sync.___internals = sync.___internals || {
    subset: function (data, select) {
      var subset = data;
      var slength = select.length;
      var skey;
      var i;
      for (i = 0; i < slength; i++) {
        skey = select[i];
        if (subset.hasOwnProperty(skey)) {
          subset = subset[skey];
        }
        else {
          return false;
        }
      }
      return subset;
    },
    item: {
      update: function (period, force_fetch) {
        var provider = this.provider;
        var fetch = sync.___internals.item.fetch;
        var render = sync.___internals.item.render;
        if (!this.ready()) {
          return;
        }
        if (typeof period === 'undefined') {
          period = this.period;
        }

        if (provider !== null) {
          if (provider.faulty === true) {
            if (period > 0) {
              // Retry when period is set.
              this.sync(this.delay + 1000, period);
            }
            return;
          }
          if (provider.fetching === true) {
            // Retry.
            this.sync(this.delay + 10, period);
            return;
          }
          if ((provider.fetched === null) || ((period > 0) && (period - 1 < Date.now() - provider.fetched)) || (force_fetch === true)) {
            fetch.call(this, period);
            return;
          }
        }

        render.call(this);
        if (period > 0) {
          // Repeat.
          this.sync(period, period);
        }
      },
      fetch: function (period) {
        var provider = this.provider;
        var fetch = sync.___internals.item.fetch;
        var render = sync.___internals.item.render;
        var request = new XMLHttpRequest();

        provider.fetching = true;
        if (typeof period === 'undefined') {
          period = this.period;
        }

        request.open('GET', provider.url, true);
        request.onload = function () {
          var data;
          if (request.status >= 200 && request.status < 400) {
            try {
              data = JSON.parse(request.responseText);
            }
            catch (e) {
              data = false;
            }
            provider.latest = data;
            provider.fetched = Date.now();
            provider.fetching = false;
            provider.faulty = false;

            render.call(this);
            if (period > 0) {
              // Repeat.
              this.sync(period, period);
            }
          }
          else {
            request.onerror();
          }
        }.bind(this);
        request.onerror = function () {
          provider.fetching = false;
          provider.faulty = true;
          if (period > 0) {
            // Retry when period is set.
            window.setTimeout(fetch.bind(this, period), period + 5000);
          }
        }.bind(this);

        request.send();
      },
      render: function () {
        var rendered;
        var el = this.element;
        var data = this.data;
        var select = this.select;
        var adjacent = this.adjacent;
        var subset = sync.___internals.subset;

        if (this.provider !== null) {
          data = this.provider.latest;
        }

        if (select !== null && typeof select === 'object' && data !== null && typeof data === 'object') {
          data = subset(data, select);
        }
        if (typeof data === 'object') {
          rendered = Mustache.render(sync.registry.templates[this.template], data);
          if (this.eval) {
            switch (adjacent) {
              case 'beforebegin':
                this.$el.before(rendered);
                break;
              case 'afterbegin':
                this.$el.prepend(rendered);
                break;
              case 'beforeend':
                this.$el.append(rendered);
                break;
              case 'afterend':
                this.$el.after(rendered);
                break;
              default:
                this.$el.html(rendered);
            }
          }
          else {
            switch (adjacent) {
              case 'beforebegin':
              case 'afterbegin':
              case 'beforeend':
              case 'afterend':
                el.insertAdjacentHTML(adjacent, rendered);
                break;
              default:
                el.innerHTML = rendered;
            }
          }
          if (el.classList) {
            el.classList.add('synced');
            el.classList.remove('syncing');
            el.classList.remove('not-synced');
          }
          else {
            el.className += ' synced';
            el.className = el.className.replace(new RegExp('(^|\\b)syncing(\\b|$)', 'gi'), ' ');
            el.className = el.className.replace(new RegExp('(^|\\b)not-synced(\\b|$)', 'gi'), ' ');
          }
        }
      },
      init: function () {
        var providers = sync.registry.providers;
        if (this.hasOwnProperty('ready')) {
          // Already initialized, aborting.
          return;
        }

        // Initialize the provider, if given.
        if (typeof this.data === 'string') {
          if (!providers.hasOwnProperty(this.data)) {
            providers[this.data] = {
              url: this.data,
              latest: {},
              fetched: null,
              fetching: false,
              faulty: false
            };
          }
          this.provider = providers[this.data];
        }
        else {
          this.data = this.data || {};
          this.provider = null;
        }

        this.ready = this.ready || sync.___internals.item.ready.bind(this);
        this.sync = this.sync || sync.___internals.item.sync.bind(this);
        this.listen = this.listen || sync.___internals.item.listen.bind(this);

        this.delay = this.delay || 0;
        this.period = this.period || 0;
        this.trigger = this.trigger || null;
        this.adjacent = this.adjacent || null;
        this.eval = this.eval || false;

        this.started = false;
        this.triggered = false;
      },
      ready: function () {
        var template_exists = sync.registry.templates.hasOwnProperty(this.template);
        if (this.eval) {
          // Script execution requires jQuery.
          if (window.jQuery && !this.$el) {
            this.$el = window.jQuery(this.element);
          }
          return template_exists && this.$el;
        }
        return template_exists;
      },
      start: function () {
        if (!this.ready() || this.started === true) {
          return;
        }
        this.started = true;

        if (this.trigger === null) {
          // Synchronize immediately, or at least after a specified delay.
          this.sync(this.delay, this.period);
        }
        else {
          // Listen and synchronize by the specified triggers.
          this.listen(this.trigger);
        }
      },
      listen: function (triggers) {
        var listeners = sync.registry.listeners;
        var trigger;
        var selector;
        var event;
        var limit;
        var group;
        var subscribers;
        var i;

        // Add this item as a subscriber for each specified triggering element.
        for (i = 0; i < triggers.length; i++) {
          trigger = triggers[i];
          selector = trigger[0];
          event = trigger[1];
          limit = trigger[2];
          if (!listeners.hasOwnProperty(selector)) {
            listeners[selector] = {
              elements: [],
              events: {}
            };
          }
          group = listeners[selector];
          if (!group.events.hasOwnProperty(event)) {
            group.events[event] = {
              triggered: function () {
                var i;
                var subscriber;
                var period;
                var length = this.subscribers.length;

                for (i = 0; i < length; i++) {
                  subscriber = this.subscribers[i];
                  period = subscriber.item.period;
                  if (!(subscriber.limit === 0)) {
                    subscriber.limit--;
                    if (subscriber.item.triggered === true) {
                      period = 0;
                    }
                    subscriber.item.triggered = true;
                    subscriber.item.sync(subscriber.item.delay, period, true);
                  }
                }
              },
              subscribers: []
            };
          }
          subscribers = group.events[event].subscribers;
          subscribers.push({item: this, limit: limit});
        }
      },
      sync: function (delay, period, force_fetch) {
        var el = this.element;
        if (typeof delay === 'undefined') {
          delay = this.delay;
        }
        if (typeof period === 'undefined') {
          period = this.period;
        }
        if (force_fetch !== true) {
          force_fetch = false;
        }
        if (el.classList) {
          el.classList.add('syncing');
          el.classList.remove('synced');
        }
        else {
          el.className += ' syncing';
          el.className = el.className.replace(new RegExp('(^|\\b)synced(\\b|$)', 'gi'), ' ');
        }
        window.setTimeout(sync.___internals.item.update.bind(this, period, force_fetch), delay);
      }
    }
  };

  sync.now = sync.now || function () {
    var registry = sync.registry;
    var init = sync.___internals.item.init;
    var start = sync.___internals.item.start;
    var item;
    var template;
    var i;

    i = sync.templates.length;
    while (i > 0) {
      i--;
      template = sync.templates.shift();
      registry.templates[template.name] = template.content;
    }

    i = registry.pending.length;
    while (i > 0) {
      i--;
      item = registry.pending.shift();
      if (item.ready()) {
        start.call(item);
      }
      else {
        registry.pending.push(item);
      }
    }

    i = sync.items.length;
    while (i > 0) {
      i--;
      item = sync.items.shift();
      init.call(item);
      registry.items.push(item);
      if (item.ready()) {
        start.call(item);
      }
      else {
        registry.pending.push(item);
      }
    }
  };
  sync.refresh = sync.refresh || function (dom) {
    var listeners = sync.registry.listeners;
    var selector;
    var group;
    var new_elements;
    var event_name;
    var event_item;
    var i;
    var el;
    var length;

    sync.now();

    if (typeof dom === 'undefined') {
      dom = window.document;
    }

    // Collect all triggering elements,
    // and register the corresponding event listeners.
    for (selector in listeners) {
      if (!listeners.hasOwnProperty(selector)) {
        continue;
      }
      group = listeners[selector];

      if (dom !== null) {
        new_elements = dom.querySelectorAll(selector);
        length = new_elements.length;
        for (i = 0; i < length; i++) {
          el = new_elements[i];
          if (group.elements.indexOf(el) < 0) {
            group.elements.push(el);
          }
          for (event_name in group.events) {
            if (!group.events.hasOwnProperty(event_name)) {
              continue;
            }
            event_item = group.events[event_name];

            length = group.elements.length;
            for (i = 0; i < length; i++) {
              group.elements[i].addEventListener(event_name, event_item.triggered.bind(event_item), false);
            }
          }
        }
      }
    }
  };

  sync.now();

  Drupal.behaviors.mustacheSync = {
    attach: function attach(dom) {
      sync.refresh(dom);
    }
  };
}(Drupal, window.mustacheSync, window.Mustache, window));
